Imports System.ComponentModel
Imports Microsoft.VisualBasic.ControlChars
Imports System.Text

Namespace System.Windows.Forms

   <System.ComponentModel.DesignerCategory("Code")> _
   Public Class FormPreferences : Inherits Component : Implements IExtenderProvider

      Private Shared _GenerateTabSize As Integer = 3
      Private _ToolStrips As New HashSet(Of ToolStrip)
      Private _NewControls As New HashSet(Of Control)

      ''' <summary>
      ''' Doppelklick im Designer auf diese Property lst Code-Generierung fr ToolstripItems aus
      ''' </summary>
      <Browsable(True)> _
      <Category("Verhalten")> _
      <DefaultValue(False)> _
      Public Property GenerateCode() As Boolean
         Get
            Return False
         End Get
         Set(ByVal value As Boolean)
            If value Then ExecuteGenerateCode(_ToolStrips)
         End Set
      End Property

      <Browsable(True)> _
      <Category("Verhalten")> _
      <DefaultValue(3)> _
      Public Property GenerateTabSize() As Integer
         Get
            Return _GenerateTabSize
         End Get
         Set(ByVal value As Integer)
            _GenerateTabSize = value
         End Set
      End Property

      Public Function CanExtend(ByVal item As Object) As Boolean Implements IExtenderProvider.CanExtend
         Dim Dum = _
         EachTime(TryCast(item, ToolStrip), AddressOf _ToolStrips.Add) OrElse _
         Once(TryCast(item, DataGridView), AddressOf DGVAction) OrElse _
         Once(TryCast(item, SplitContainer), AddressOf SplitContainerAction) OrElse _
         Once(TryCast(item, Label), Function(Ctl) Ctl.AutoSize.Assign(False)) OrElse _
         Once(TryCast(item, TreeView), Function(Ctl) Ctl.HideSelection.Assign(False)) OrElse _
         Once(TryCast(item, TextBoxBase), Function(Ctl) Ctl.HideSelection.Assign(False)) OrElse _
         Once(TryCast(item, ListView), Function(Ctl) Ctl.HideSelection.Assign(False))
      End Function

      Private Function Once(Of T As Control)(ByVal Ctl As T, ByVal Preference As Action(Of T)) As Boolean
         'Nur fr **neue** Controls drfen die Preferenzen ausgefhrt werden.
         'neue Controls werden durch Test auf .Parent=Nothing identifiziert.
         'leider ist fr ToolstripItem keine Preferenz definierbar, da nicht von Control erbend.
         Once = Ctl IsNot Nothing
         If Once Then
            If Ctl.Parent Is Nothing Then
               _NewControls.Add(Ctl)
            Else
               If _NewControls.Remove(Ctl) Then Preference(Ctl)
            End If
         End If
      End Function

      Private Function EachTime(Of T)(ByVal Obj As T, ByVal Preference As Action(Of T)) As Boolean
         EachTime = Obj IsNot Nothing
         If EachTime Then Preference(Obj)
      End Function

      Private Sub SplitContainerAction(ByVal spl As SplitContainer)
         spl.SplitterWidth = 8
         spl.FixedPanel = FixedPanel.Panel1
      End Sub

      Private Sub DGVAction(ByVal dgv As DataGridView)
         dgv.RowHeadersWidth = 25
         dgv.AllowUserToResizeRows = False
         dgv.AutoSizeColumnsMode = DataGridViewAutoSizeColumnsMode.DisplayedCells
      End Sub

#Region "ExecuteGenerateCode"

      Public Shared Sub ExecuteGenerateCode(ByVal _ToolStrips As HashSet(Of ToolStrip))
         Dim AllWriter As New StringList
         Dim Clickables As New StringList
         For Each CustomerStrip In _ToolStrips
            Clickables.Clear()
            For Each Itm As ToolStripItem In CustomerStrip.Items
               CollectClickables(Itm, Clickables)
            Next
            If Clickables.Count > 0 Then
               Dim sHandles = Clickables.SetDelimiter(".Click, ").ToString & ".Click"
               Dim sCase = Tab & Tab & "case Sender Is "
               Dim sCases = sCase & Clickables.SetDelimiter(CrLf & CrLf & sCase).ToString & CrLf
               With AllWriter
                  .AddLine("Private Sub ", CustomerStrip.Name, "_MenuClicked(ByVal Sender As Object, ByVal e As EventArgs) _")
                  .AddLine(Tab, Tab, "Handles ", sHandles)
                  .AddLine(Tab, "select case True")
                  .AddLine(sCases)
                  .AddLine(Tab, "End select")
                  .AddLine("End Sub")
                  .AddLine("")
               End With
            End If
         Next
         'Die Ausgabe wird in eine TextDatei geschrieben, die dann geffnet wird
         Dim S = AllWriter.SetDelimiter("").ToString.Replace(Tab, New String(" "c, _GenerateTabSize))
         Dim FileFullName As String = Environment.CurrentDirectory & "\LaunchEditorWith.txt"
         My.Computer.FileSystem.WriteAllText(FileFullName, S, False)
         System.Diagnostics.Process.Start(FileFullName).Dispose()
      End Sub

      Private Shared Sub CollectClickables(ByVal Itm As ToolStripItem, ByVal Clickables As StringList)
         If TypeOf Itm Is ToolStripDropDownItem Then
            With DirectCast(Itm, ToolStripDropDownItem)
               If .HasDropDownItems Then
                  For Each itmRecurse As ToolStripItem In .DropDownItems
                     CollectClickables(itmRecurse, Clickables)        ' Rekursion!
                  Next
                  Return
               End If
            End With
            Clickables.Add(Itm.Name)
         ElseIf TypeOf Itm Is ToolStripButton Then
            Clickables.Add(Itm.Name)
         End If
      End Sub

#End Region 'ExecuteGenerateCode

   End Class

End Namespace