﻿Imports System.Runtime.CompilerServices
Imports System.Reflection
Imports System.ComponentModel
Imports System.Data
Imports System.IO

Namespace System.Windows.Forms
   Public Module FormX
      Private _TableAttachs As New ComponentProperty(Of DataTable, DataTableAttach)(Function(tb) New DataTableAttach(tb))
      <Extension()> _
      Public Function Attach(ByVal tb As DataTable) As DataTableAttach
         Return _TableAttachs(tb)
      End Function

      <Extension()> _
      Private Function Attach(ByVal dts As DataSet) As DatasetAdapterBase
         Return DatasetAdapterBase.Attachs(dts)
      End Function
      <Extension()> _
      Public Function DataFile(ByVal dts As DataSet, ByVal path As String) As DataSet
         DatasetAdapterBase.Attachs(dts) = New DatasetXmlAdapter(dts, path)
         Return dts
      End Function
      <Extension()> _
      Public Function DataFile(ByVal dts As DataSet) As FileInfo
         Dim att = DirectCast(dts.Attach, DatasetXmlAdapter)
         If att.Null Then Return Nothing
         Return att.DataFile
      End Function

      <Extension()> _
      Public Function LoadXml(ByVal dts As DataSet) As DataSet
         DirectCast(dts.Attach, DatasetXmlAdapter).Fill()
         Return dts
      End Function
      <Extension()> _
      Public Sub Save(ByVal dts As DataSet, ByVal frm As Form, Optional ByVal checkForChanges As Boolean = True)
         dts.Attach.Save(frm, checkForChanges)
      End Sub

#Region "evtl. obsolet - erst noch drinnelassen"
      'für eine ProgrammArchitektur mit Sub Main wären diese hier nützlich gwä.

      Public Sub PrepareApp()
         Application.EnableVisualStyles()
         Application.SetCompatibleTextRenderingDefault(False)
      End Sub
      <Extension()> _
      Public Sub RunApplication(ByVal mainForm As Form)
         RemoveHandler mainForm.FormClosed, AddressOf Form_Closed
         AddHandler mainForm.FormClosed, AddressOf MainForm_Closed
         Application.Run(mainForm)
         My.Settings.Save()
      End Sub

#End Region 'evtl. obsolet - erst noch drinnelassen

      ''' <summary>dockt das Form am oberern Bildschirmrand, in voller Breite</summary>
      <Extension()> _
      Public Function AlignOnTop(Of T As Form)(ByVal frm As T) As T
         AlignOnTop = frm
         If frm.WindowState = FormWindowState.Maximized Then Exit Function
         Dim rct = Screen.PrimaryScreen.WorkingArea
         rct.Height = frm.Height
         frm.StartPosition = FormStartPosition.Manual
         frm.Bounds = rct
      End Function
#Region "Winfom-Close-Bug"      ' Form hat den Bug, dasses die Komponenten vor den Controls disposet. Das ergibt Datagridview-Fehler, wenn im DGV ComboboxColumns verbaut sind, die an BindingSource-Komponenten gebunden sind. Hiermit wern also die Controls zuerst weggehauen - dass da nixme anbrennt

      Private Sub Form_Closed(ByVal s As Object, ByVal e As EventArgs)
         For Each ctl As Control In DirectCast(s, Form).Controls
            ctl.Dispose()
         Next
      End Sub

      'Beim Schließen des MainForms wird Form_Closed der anneren Forms garnet mehr aufgerufen. Daher hier alle alle Controls disposen
      Private Sub MainForm_Closed(ByVal s As Object, ByVal e As EventArgs)
         For Each frm As Form In Application.OpenForms
            For Each ctl As Control In frm.Controls
               ctl.Dispose()
            Next
         Next
      End Sub

#End Region 'Winfom-Close-Bug
      <Extension()> _
      Public Function HandleFormClosing(ByVal dts As DataSet) As FormClosingEventHandler
         Return dts.Attach.HandleFormClosing
      End Function

      Private _KnownDatasets As New Dictionary(Of Type, DataSet)

      <Extension()> _
      Public Function Register(ByVal dts As DataSet, ByVal frm As Form) As DataSet
         Dim tp = dts.GetType
         Dim ds As DataSet = Nothing
         If Not _KnownDatasets.TryGetValue(tp, ds) Then
            ds = dts
            _KnownDatasets.Add(tp, ds)
         End If
         frm.SetDataSet(ds)
         Return ds
      End Function

      <Extension()> _
      Public Function SetDataSet(Of T As Form)(ByVal frm As T, ByVal dts As DataSet) As T
         SetDataSet = frm
         AddHandler frm.FormClosed, AddressOf Form_Closed
         Dim allCtls = New GetChilds(Of Control)(Function(ctl) ctl.Controls).AllAsList(frm)
         Dim dgvErrorSuspend = New GridErrorSuspend(dts, allCtls)
         Dim tpDts = dts.GetType
         Dim oldDatasets As New HashSet(Of DataSet)
         Dim shouldReplace As Func(Of Object, Boolean) = Function(src) _
            src IsNot Nothing AndAlso _
            (src Is tpDts OrElse src.GetType Is tpDts AndAlso src IsNot dts)
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         For Each ctl In allCtls.OfType(Of ContainerControl)()
            For Each fld In ctl.GetType.GetFields(bindFlags)
               If fld.FieldType Is tpDts Then
                  oldDatasets.Add(DirectCast(fld.GetValue(ctl), DataSet))
                  fld.SetValue(ctl, dts)
                  Continue For
               End If
               If Not GetType(IComponent).IsAssignableFrom(fld.FieldType) Then Continue For
               Dim prp = fld.FieldType.GetProperty("DataSource")
               If prp.Null Then Continue For
               Dim itm = fld.GetValue(ctl)
               If itm Is Nothing Then Continue For
               Dim bs = TryCast(itm, BindingSource)
               If bs.NotNull Then
                  Dim tb = bs.DataTable
                  If tb.NotNull Then
                     tb = dts.Tables(tb.TableName)
                     If tb.NotNull Then tb.Attach.AddSource(bs)
                  End If
               End If
               If shouldReplace(prp.GetValue(itm)) Then prp.SetValue(itm, dts)
            Next
         Next
         dgvErrorSuspend.Resume()
         oldDatasets.Remove(dts)
         For Each dts In oldDatasets : dts.Dispose() : Next
      End Function

      Private Class GridErrorSuspend
         'beim Umstöpseln der DataSources können in DGVs DataErrors auftreten. Diese werden suspendiert, wenn temporär ein  DataGridViewDataErrorEventHandler das dgv.DataError-Event behandelt 
         Private _Grids As New List(Of DataGridView)
         Private _DummiHandler As DataGridViewDataErrorEventHandler = Function() 0
         Public Sub New(ByVal dts As DataSet, ByVal allCtls As List(Of Control))
            If dts.Tables.Cast(Of DataTable).All(Function(tb) tb.Rows.Count = 0) Then Return
            For Each ctl In allCtls
               Dim dgv = TryCast(ctl, DataGridView)
               If dgv.Null Then Continue For
               AddHandler dgv.DataError, _DummiHandler
               _Grids.Add(dgv)
            Next
         End Sub
         Public Sub [Resume]()
            For Each dgv In _Grids
               RemoveHandler dgv.DataError, _DummiHandler
               Dim bs = TryCast(dgv.DataSource, BindingSource)
               If bs.NotNull Then bs.ResetBindings(False)
            Next
         End Sub
      End Class
   End Module
End Namespace

