﻿Imports System.Data
Imports System.Windows.Forms
Imports System.Data.Common
Imports System.ComponentModel

'2 Möglichkeiten, einen DatasetAdapter zu erzeugen: 
'1) DatasetAdapter.FromTableAdapters() bohrt per Reflection die anzugebenden generierten TableAdapter auf, und verwendet die darin verkapselten DataAdapter
'2) DatasetAdapter.FromTables() creiert die erf. DataAdapter mittels CommandBuilder

Namespace System.Data

   Partial Public Class DatasetAdapter

      'zwSpeicher, die erst im EndInit verarbeitet werden
      Private _InitByTables As DataTable()() = New DataTable()() {New DataTable() {}, New DataTable() {}}
      Private _InitByAdapters As Component()() = New Component()() {New Component() {}, New Component() {}}
      Private _InitNoFill(-1) As DataTable
      Private _InitOption As ConflictOption

      Private Sub New()
      End Sub

#Region "InitFromAdapters"

      Public Shared Function FromTableAdapters(ByVal dataset As DataSet) As IInitFromTableAdapters
         Return New DatasetAdapter With {._Dataset = dataset}
      End Function

      Public Interface IInitFromTableAdapters
         'definiert das Flow-interface: Um einen DatasetAdapter zu initialisieren muß man an DatasetAdapter.Create() das Dataset übergeben, und dann für jede DataTable einen generierten TableAdapter angeben.
         'Dabei wird zw. Adapters und IncrementalAdapters unterschieden. Letztere werden nämlich nicht beim Load() befüllt, sondern durch spezifischer Aufrufe, meist durch FillByParentRow().
         'FillByParentRow(table, parentRow) befüllt die ZielTabelle nur mit den der angegebenen Parentrow untergeordneten Datensätzen. Dabei kann die parentRow sogar über mehrere Zwischenstufen übergeordnet sein.
         'FillFromWhere(DataTable, String, Object()) füllt die angegebene Tabelle, wobei FROM- und WHERE- Segment frei definierbar sind, und Parameter angegeben werden können.

         ''' <summary> geben Sie hier TableAdapter an, die "ihre" DataTable komplett laden sollen </summary>
         Function CompleteFill(ByVal ParamArray GeneratedAdapters As Component()) As IInitFromTableAdapters

         ''' <summary> geben Sie hier TableAdapter an, die "ihre" DataTable nur in Abhängigkeit von
         '''  übergeordneten DataRows befüllen sollen </summary>
         Function IncrementalFill(ByVal ParamArray GeneratedAdapters As Component()) As IInitFromTableAdapters

         ''' <summary> geben Sie hier DataTables an, die nicht befüllt werden sollen </summary>
         Function NoFill(ByVal ParamArray tables As DataTable()) As IInitFromTableAdapters

         ''' <summary> schließt die Initialisierung ab </summary>
         Function EndInit() As DatasetAdapter
      End Interface

      Private Function CompleteFill(ByVal ParamArray GeneratedAdapters() As Component) As IInitFromTableAdapters Implements IInitFromTableAdapters.CompleteFill
         _InitByAdapters(0) = GeneratedAdapters
         Return Me
      End Function

      Private Function IncrementalFill(ByVal ParamArray GeneratedAdapters() As Component) As IInitFromTableAdapters Implements IInitFromTableAdapters.IncrementalFill
         _InitByAdapters(1) = GeneratedAdapters
         Return Me
      End Function

      Private Function NoFill(ByVal ParamArray tables() As System.Data.DataTable) As IInitFromTableAdapters Implements IInitFromTableAdapters.NoFill
         _InitNoFill = tables
         Return Me
      End Function

      Private Function EndInitFromAdapters() As DatasetAdapter Implements IInitFromTableAdapters.EndInit
         Dim converter As Converter(Of Component, TableAdapter) _
            = Function(a) TableAdapter.FromCrackedAdapter(a, _Dataset)
         Return EndInit(_InitByAdapters(0).ConvertAll(converter), _InitByAdapters(1).ConvertAll(converter))
      End Function

#End Region 'InitFromAdapters

#Region "InitFromTables"

      Public Shared Function FromTables(ByVal con As DbConnection, ByVal conflictOption As ConflictOption) As IInitFromTables
         con.Open() 'für schnelle CommandBuilder-Zugriffe - wird in EndInitFromTables() geschlossen
         Return New DatasetAdapter With {._Connection = con, ._InitOption = conflictOption}
      End Function

      Public Interface IInitFromTables
         ' Flow-interface: Um einen DatasetAdapter zu initialisieren muß man an DatasetAdapter.Create() das Dataset übergeben, und dann für jede DataTable einen generierten TableAdapter angeben.
         'Dabei wird zw. Adapters und IncrementalAdapters unterschieden. Letztere werden nämlich nicht beim Load() befüllt, sondern durch spezifischer Aufrufe, meist durch FillByParentRow().
         'FillByParentRow(table, parentRow) befüllt die ZielTabelle nur mit den der angegebenen Parentrow untergeordneten Datensätzen. Dabei kann die parentRow sogar über mehrere Zwischenstufen übergeordnet sein.
         'FillFromWhere(DataTable, String, Object()) füllt die angegebene Tabelle, wobei FROM- und WHERE- Segment frei definierbar sind, und Parameter angegeben werden können.

         ''' <summary> geben Sie hier DataTables an, die komplett laden sollen </summary>
         Function CompleteFill(ByVal ParamArray tables As DataTable()) As IInitFromTables

         ''' <summary> geben Sie hier DataTables an, die nur in Abhängigkeit von
         '''  übergeordneten DataRows befüllen werden sollen </summary>
         Function IncrementalFill(ByVal ParamArray tables As DataTable()) As IInitFromTables

         ''' <summary> geben Sie hier DataTables an, die nicht befüllt werden sollen </summary>
         Function NoFill(ByVal ParamArray tables As DataTable()) As IInitFromTables

         ''' <summary> schließt die Initialisierung ab </summary>
         Function EndInit() As DatasetAdapter
      End Interface

      Private Function CompleteFill(ByVal ParamArray tables() As DataTable) As IInitFromTables Implements IInitFromTables.CompleteFill
         _InitByTables(0) = tables
         Return Me
      End Function

      Private Function IncrementalFill(ByVal ParamArray tables() As DataTable) As IInitFromTables Implements IInitFromTables.IncrementalFill
         _InitByTables(1) = tables
         Return Me
      End Function

      Private Function NoFillFromTables(ByVal ParamArray tables() As DataTable) As IInitFromTables Implements IInitFromTables.NoFill
         _InitNoFill = tables
         Return Me
      End Function

      Private Function EndInitFromTables() As DatasetAdapter Implements IInitFromTables.EndInit
         Try
            Dim converter As Converter(Of DataTable, TableAdapter) _
               = Function(tb As DataTable) TableAdapter.FromTable(tb, _Connection, _InitOption)
            Return EndInit(_InitByTables(0).ConvertAll(converter), _InitByTables(1).ConvertAll(converter))
         Finally
            _Connection.Close()
         End Try
      End Function

#End Region 'InitFromTables

#Region "common Initfunctions"

      Private Function EndInit(ByVal fillAdapters As TableAdapter(), ByVal incrementAdapters As TableAdapter()) As DatasetAdapter
         _FillAdapters = TopoSort(fillAdapters)
         _IncrementAdapters = TopoSort(incrementAdapters)
         _UpdateAdapters = _FillAdapters.Concat(_IncrementAdapters).ToArray
         Validate()
         InitAutoValues(_FillAdapters, False)
         InitAutoValues(_IncrementAdapters, True)
         _InitByTables = Nothing
         _InitByAdapters = Nothing
         _InitNoFill = Nothing
         Return Me
      End Function

      Private Function TopoSort(ByVal adps As TableAdapter()) As List(Of TableAdapter)
         'DB-Zugriffe verknüpfter Tabellen erfordern, dass die übergeordneten Tabellen zuerst bearbeitet werden. Mathematisch gesehen handelt es sich um die topologische Sortierung des Graphen "Dataset"
         TopoSort = New List(Of TableAdapter)
         For Each adp In adps
            'topologisch sortiert: ParentTables vor ihren Kindern ordnen
            Dim tb = adp.Table
            TopoSort.InsertOrAdd(adp, Function(a) tb.IsParentOf(a.Table))
         Next
      End Function

      ''' <summary>richtet Generierung von Primärschlüsselwerten beim Client ein</summary>
      ''' <remarks>
      ''' unterstützt Datentypen, die auf AutoIncrement gesetzt werden können, sowie System.Guid
      ''' </remarks>
      Private Sub InitAutoValues(ByVal adps As IList(Of TableAdapter), ByVal getIncrementSeed As Boolean)
         For Each adp In adps
            Dim tb = adp.Table
            Dim PrimKey As DataColumn() = tb.PrimaryKey
            If PrimKey.Length <> 1 Then Continue For
            Dim primCol As DataColumn = PrimKey(0)
            Dim tp As Type = primCol.DataType
            If tp.Equals(GetType(Int16)) OrElse _
                  tp.Equals(GetType(Int32)) OrElse _
                  tp.Equals(GetType(Int64)) Then
               primCol.AutoIncrement = True
               primCol.AutoIncrementStep = 1
               primCol.AutoIncrementSeed = 1
               If getIncrementSeed Then
                  Using cmd = Connection.CreateCommand, eo = Connection.EnterOpen
                     cmd.CommandText.Set("SELECT MAX(", primCol.ColumnName, ") FROM ", tb.TableName, ";")
                     Dim val = cmd.ExecuteScalar()
                     primCol.AutoIncrementSeed = If(Convert.IsDBNull(val), 0, CInt(val)) + 1
                  End Using
               End If
            ElseIf tp.Equals(GetType(Guid)) Then
               AddHandler primCol.Table.TableNewRow, AddressOf Table_NewRow
            End If
         Next
      End Sub

      Private Sub Table_NewRow(ByVal sender As Object, ByVal e As DataTableNewRowEventArgs)
         e.Row(e.Row.Table.PrimaryKey(0)) = Guid.NewGuid
      End Sub

      Private Sub Validate()
         If _UpdateAdapters.Count = 0 Then
#If DEBUG Then
            Stop
#End If
            Throw Me.Exception("Es wurden keine TableAdapter initialisiert")
         End If
         _Dataset = _UpdateAdapters.First.Table.DataSet
			If _InitNoFill.Length > 0 AndAlso _UpdateAdapters.Count + _InitNoFill.Length <> _Dataset.Tables.Count Then
#If DEBUG Then
				Stop
#End If
				Throw Me.Exception( _
				 "Für ", _UpdateAdapters.Count, " DataTables wurden TableAdapter angegeben.", _
				 " ", _InitNoFill.Length, " Tables sollen nicht persistiert werden.", _
				 "Es sind aber ", _Dataset.Tables.Count, " Tables im Dataset enthalten")
			End If
         Dim fillParents = _FillAdapters _
            .SelectMany(Function(a) a.Table.ParentRelations.Cast(Of DataRelation) _
               .Select(Function(rl) rl.ParentTable)).Distinct
         If _IncrementAdapters.Any(Function(a) fillParents.Contains(a.Table)) Then
#If DEBUG Then
            Stop
#End If
            Throw Me.Exception( _
               "Unter den vollständig zu befüllenden Tables finden sich welche, deren ParentTable als incrementell zu befüllen angegeben wurde. Das wird schief gehen.")
         End If
      End Sub

      Public Sub RegisterMainSource(ByVal mainSource As BindingSource)
         Dim sources = mainSource.Container.Components.OfType(Of BindingSource)().Except(mainSource).ToArray
         For Each bs In sources
            If TypeOf bs.DataSource Is DataSet Then bs.DataSource = mainSource
         Next
         mainSource.DataSource = _Dataset
         For Each bs In sources
            bs.SubscribeEventDisabling()
            _SourceInfos.Add(bs)
         Next
      End Sub

#End Region 'common Initfunctions

   End Class

End Namespace